# Copyright 2026 Rigetti & Co, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from functools import reduce
from operator import mul
from typing import TYPE_CHECKING, Any, Iterator, Self, Tuple

import jax
import jax.numpy as jnp
from jax import Array

if TYPE_CHECKING:
    import qutip
    from numpy.typing import NDArray


# ---------- base mixin ----------


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class State:
    """Base class for a quantum state."""

    data: Array
    """The state tensor with shape (*ensemble, d0, d1, ...) for state vectors or
    (*ensemble, d0_out, d1_out, ..., d0_in, d1_in, ...) for density matrices."""

    num_ensemble_dims: int = 0
    """The number of leading ensemble dimensions in the data array."""

    def __neg__(self) -> Self:
        """Multiply the unitary by -1."""
        return type(self)(-self.data, self.num_ensemble_dims)

    def __mul__(self, scalar: complex) -> Self:
        """Scalar multiplication of the superoperator."""
        return type(self)(self.data * scalar, self.num_ensemble_dims)

    def __rmul__(self, scalar: complex) -> Self:
        """Scalar multiplication of the superoperator."""
        return self * scalar

    def __pow__(self, exponent: float) -> Self:
        """Exponentiation of the state"""
        raise NotImplementedError(f"Exponentiation not implemented for {type(self)}.")

    def __str__(self) -> str:
        if self.ensemble_size != ():
            return (
                f"{type(self).__name__}(dims={self.dims}, ensemble_size={self.ensemble_size}, shape={self.data.shape})"
            )
        return f"{type(self).__name__}(dims={self.dims}, shape={self.data.shape})"

    def __eq__(self, other: Any) -> bool:
        """Check equality of two Operators."""
        # This eq method will never be _wrong_ but some operators are equivalent
        # up to a global phase or other transformations
        if not isinstance(other, type(self)):
            return False
        if self.dims != other.dims:
            return False
        elif jnp.allclose(self.data, other.data):
            return True
        else:
            return False

    def __matmul__(self, other: Any) -> Any:
        """Matrix multiplication of the operator with another operator."""
        # composition is very different between different operator types and should be delegated
        raise NotImplementedError(f"Matrix multiplication not implemented between {type(self)} and {type(other)}.")

    def __or__(self, other: Any) -> Any:
        """Tensor product of the operator with another operator."""
        # tensor product is very differentse between different operator types and should be delegated
        raise NotImplementedError(f"Tensor product not implemented between {type(self)} and {type(other)}.")

    def tree_flatten(self):
        return (self.data,), self.num_ensemble_dims

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        (data,) = children
        return cls(data=data, num_ensemble_dims=aux_data)

    def conj(self):
        """Complex conjugate of the states(s)"""
        return type(self)(data=jnp.conjugate(self.data), num_ensemble_dims=self.num_ensemble_dims)

    @property
    def T(self):
        """Transpose of the operator(s)"""
        raise NotImplementedError(f"Transpose not implemented for {type(self)}.")

    @property
    def h(self):
        """Hermitian conjugate of the operator(s)"""
        raise NotImplementedError(f"Hermitian conjugate not implemented for {type(self)}.")

    @property
    def dims(self) -> Tuple[int, ...]:
        """The dimensions of each qudit, inferred from data shape."""
        raise NotImplementedError("Subclasses must implement dims property.")

    @property
    def d(self) -> int:
        return reduce(mul, self.dims)

    @property
    def d2(self) -> int:
        return self.d**2

    @property
    def ensemble_size(self) -> Tuple[int, ...]:
        """Returns the size of the ensemble if the state represents an ensemble of states."""
        raise NotImplementedError("Subclasses must implement ensemble_size property.")

    @property
    def matrix(self) -> Array:
        """Returns the matrix representation of the state."""
        raise NotImplementedError("Subclasses must implement matrix property.")

    @classmethod
    def from_matrix(cls, matrix: Array, dims: Tuple[int, ...], num_ensemble_dims: int = 0) -> Self:
        """Construct from matrix representation."""
        raise NotImplementedError("Subclasses must implement from_matrix class method.")


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class Operator:
    """Base class for a quantum operator."""

    data: Array
    """The operator tensor with shape (*ensemble, d0_out, d1_out, ..., d0_in, d1_in, ...)."""

    num_ensemble_dims: int = 0
    """The number of leading ensemble dimensions in the data array."""

    def __neg__(self) -> Self:
        """Multiply the unitary by -1."""
        return type(self)(-self.data, self.num_ensemble_dims)

    def __mul__(self, scalar: complex) -> Self:
        """Scalar multiplication of the superoperator."""
        return type(self)(self.data * scalar, self.num_ensemble_dims)

    def __rmul__(self, scalar: complex) -> Self:
        """Scalar multiplication of the superoperator."""
        return self * scalar

    def __pow__(self, exponent: float) -> Self:
        """Exponentiation of the operator"""
        # By default, only support integer exponents
        # matrix_power is not _correct_ for all operators
        if jnp.isclose(int(jnp.abs(exponent)), exponent):
            new_data = jnp.linalg.matrix_power(self.matrix, int(exponent))
            return type(self).from_matrix(new_data, self.dims, self.num_ensemble_dims)
        else:
            raise TypeError(f"Exponent must be an integer, but got {type(exponent)}.")

    def __str__(self) -> str:
        if self.ensemble_size != ():
            return (
                f"{type(self).__name__}(dims={self.dims}, ensemble_size={self.ensemble_size}, shape={self.data.shape})"
            )
        return f"{type(self).__name__}(dims={self.dims}, shape={self.data.shape})"

    def __eq__(self, other: Any) -> bool:
        """Check equality of two Operators."""
        # This eq method will never be _wrong_ but some operators are equivalent
        # up to a global phase or other transformations
        if not isinstance(other, type(self)):
            return False
        if self.dims != other.dims:
            return False
        elif jnp.allclose(self.data, other.data):
            return True
        else:
            return False

    def __rmatmul__(self, other: Any) -> Any:
        """Matrix multiplication of the operator with another operator."""
        # composition is very different between different operator types and should be delegated
        raise NotImplementedError(f"Matrix multiplication not implemented between {type(self)} and {type(other)}.")

    def __or__(self, other: Any) -> Any:
        """Tensor product of the operator with another operator."""
        # tensor product is very different between different operator types and should be delegated
        raise NotImplementedError(f"Tensor product not implemented between {type(self)} and {type(other)}.")

    def tree_flatten(self):
        return (self.data,), self.num_ensemble_dims

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        (data,) = children
        return cls(data=data, num_ensemble_dims=aux_data)

    def conj(self):
        """Complex conjugate of the operator(s)"""
        return type(self)(data=jnp.conjugate(self.data), num_ensemble_dims=self.num_ensemble_dims)

    @property
    def dims(self) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """The (output, input) dimensions of each qudit, inferred from data shape."""
        raise NotImplementedError("Subclasses must implement dims property.")

    @property
    def T(self):
        """Transpose of the operator(s)"""
        # use the swapaxes function to swap the last two axes
        matrix_t = jnp.swapaxes(self.matrix, -1, -2)
        return type(self).from_matrix(matrix_t, (self.dims[1], self.dims[0]), self.num_ensemble_dims)

    @property
    def h(self):
        """Hermitian conjugate of the operator(s)"""
        matrix_h = jnp.conjugate(jnp.swapaxes(self.matrix, -1, -2))
        return type(self).from_matrix(matrix_h, (self.dims[1], self.dims[0]), self.num_ensemble_dims)

    @property
    def d(self) -> Tuple[int, int]:
        return tuple(reduce(mul, dim) for dim in self.dims)

    @property
    def d2(self) -> Tuple[int, int]:
        return self.d[0] ** 2, self.d[1] ** 2

    @property
    def ensemble_size(self) -> Tuple[int, ...]:
        """Returns the size of the ensemble if the operator represents an ensemble of operators."""
        return self.data.shape[: self.num_ensemble_dims]

    @property
    def matrix(self) -> Array:
        """Returns the matrix representation (*ensemble, d_out, d_in) of the operator."""
        ensemble_shape = self.data.shape[: self.num_ensemble_dims]
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        n_qudits = len(qudit_shape) // 2
        d_out = reduce(mul, qudit_shape[:n_qudits], 1)
        d_in = reduce(mul, qudit_shape[n_qudits:], 1)
        return self.data.reshape(ensemble_shape + (d_out, d_in))

    @classmethod
    def from_matrix(
        cls, matrix: Array, dims: Tuple[Tuple[int, ...], Tuple[int, ...]], num_ensemble_dims: int = 0
    ) -> Self:
        """Construct from matrix representation.

        :param matrix: Array with shape (*ensemble, d_out, d_in)
        :param dims: Tuple of (dims_out, dims_in) where each is a tuple of qudit dimensions
        :param num_ensemble_dims: Number of leading ensemble dimensions
        :return: Operator with tensor data
        """
        ensemble_shape = matrix.shape[:num_ensemble_dims]
        tensor = matrix.reshape(ensemble_shape + dims[0] + dims[1])
        return cls(data=tensor, num_ensemble_dims=num_ensemble_dims)

    def __iter__(self) -> Iterator[Self]:
        if self.num_ensemble_dims == 0:
            raise TypeError(
                "This Operator is not ensembled (num_ensemble_dims == 0), so it cannot be iterated. "
                "If you intended an ensemble, make data shape (N, d0_out, d1_out, ..., d0_in, d1_in, ...) (or higher-rank)."
            )
        # iterate over axis 0 only; remaining ensemble axes (if any) remain on each item
        for i in range(self.data.shape[0]):
            yield type(self)(data=self.data[i], num_ensemble_dims=self.num_ensemble_dims - 1)


# This class is basically for typing purposes
# Some methods works on any sort of Superoperator
@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class SuperOperator(Operator):
    """Base class for a quantum superoperator.

    SuperOperators have a 4-group tensor structure:
    (*ensemble, d0_out_bra, d1_out_bra, ..., d0_out_ket, d1_out_ket, ...,
            d0_in_bra, d1_in_bra, ..., d0_in_ket, d1_in_ket, ...)
    """

    @property
    def matrix(self) -> Array:
        """Returns the matrix representation (*ensemble, d_out^2, d_in^2) of the superoperator."""
        ensemble_shape = self.data.shape[: self.num_ensemble_dims]
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        # 4 groups of dimensions: out_bra, out_ket, in_bra, in_ket
        n_qudits = len(qudit_shape) // 4
        d_out_bra = reduce(mul, qudit_shape[:n_qudits], 1)
        d_out_ket = reduce(mul, qudit_shape[n_qudits : 2 * n_qudits], 1)
        d_in_bra = reduce(mul, qudit_shape[2 * n_qudits : 3 * n_qudits], 1)
        d_in_ket = reduce(mul, qudit_shape[3 * n_qudits :], 1)
        d_out = d_out_bra * d_out_ket
        d_in = d_in_bra * d_in_ket
        return self.data.reshape(ensemble_shape + (d_out, d_in))

    @classmethod
    def from_matrix(
        cls, matrix: Array, dims: Tuple[Tuple[int, ...], Tuple[int, ...]], num_ensemble_dims: int = 0
    ) -> Self:
        """Construct from matrix representation.

        :param matrix: Array with shape (*ensemble, d_out^2, d_in^2)
        :param dims: Tuple of (dims_out, dims_in) where each is a tuple of qudit dimensions
        :param num_ensemble_dims: Number of leading ensemble dimensions
        :return: SuperOperator with tensor data
        """
        ensemble_shape = matrix.shape[:num_ensemble_dims]
        # Tensor shape is: out_bra_dims + out_ket_dims + in_bra_dims + in_ket_dims
        tensor_shape = dims[0] + dims[0] + dims[1] + dims[1]
        tensor = matrix.reshape(ensemble_shape + tensor_shape)
        return cls(data=tensor, num_ensemble_dims=num_ensemble_dims)


# ---------- states ----------


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class StateVector(State):
    """State vector |psi>, shape (*ensemble, d0, d1, ...) in tensor form or (*ensemble, d) in matrix form."""

    @property
    def dims(self) -> Tuple[int, ...]:
        """The dimensions of each qudit, inferred from data shape."""
        return self.data.shape[self.num_ensemble_dims :]

    @property
    def ensemble_size(self) -> Tuple[int, ...]:
        """Returns the size of the ensemble if the state represents an ensemble of states."""
        return self.data.shape[: self.num_ensemble_dims]

    @property
    def matrix(self) -> Array:
        """Returns the vector representation (*ensemble, d) of the state."""
        ensemble_shape = self.data.shape[: self.num_ensemble_dims]
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        d = reduce(mul, qudit_shape, 1)
        return self.data.reshape(ensemble_shape + (d,))

    @classmethod
    def from_matrix(cls, matrix: Array, dims: Tuple[int, ...], num_ensemble_dims: int = 0) -> "StateVector":
        """Construct from vector representation.

        :param matrix: Array with shape (*ensemble, d) where d = prod(dims)
        :param dims: Tuple of qudit dimensions (d0, d1, ...)
        :param num_ensemble_dims: Number of leading ensemble dimensions
        :return: StateVector with tensor data
        """
        ensemble_shape = matrix.shape[:num_ensemble_dims]
        tensor = matrix.reshape(ensemble_shape + dims)
        return cls(data=tensor, num_ensemble_dims=num_ensemble_dims)

    @property
    def T(self):
        """Transpose of the operator(s)"""
        # The state vector is a 1D array, so transpose does nothing
        return self

    @property
    def h(self):
        """Hermitian conjugate of the operator(s)"""
        # The state vector is a 1D array, so hermitian conjugate is just complex conjugate
        return self.conj()

    def __matmul__(self, other):
        """Left multiply the state by another."""
        match other:
            case StateVector():  # <𝜓|𝜙> -> p
                return jnp.einsum("...a,...a->...", self.matrix.conj(), other.matrix)
            case DensityMatrix():  #  <𝜓|𝜌 -> <𝜙|
                result = jnp.einsum("...b,...ba->...a", self.matrix.conj(), other.matrix)
                num_ensemble = max(self.num_ensemble_dims, other.num_ensemble_dims)
                return StateVector.from_matrix(result, self.dims, num_ensemble)
            case _:
                return NotImplemented

    def __or__(self, other: Any) -> Any:
        """Tensor product of the state vector with another state."""
        match other:
            case StateVector():
                # |ψ1⟩ ⊗ |ψ2⟩ -> StateVector
                from ._tensor import tensor_state_vector

                return tensor_state_vector(self, other)
            case DensityMatrix():
                # |ψ⟩ ⊗ ρ -> DensityMatrix (promote state vector)
                from ._promotion import promote_state_vector_to_density_matrix
                from ._tensor import tensor_density_matrix

                return tensor_density_matrix(promote_state_vector_to_density_matrix(self), other)
            case Unitary() | SuperOp() | Choi() | PauliLiouville() | KrausMap():
                # State | Operator -> NotImplemented
                return NotImplemented
            case _:
                return NotImplemented

    def __eq__(self, other: Any) -> bool:
        """Check equality using quantum state fidelity."""
        match other:
            case StateVector():
                # Compare two state vectors using fidelity
                from ._distance_metrics import fidelity

                return bool(jnp.allclose(fidelity(self, other), 1.0))
            case DensityMatrix():
                # Promote self to density matrix and compare
                from ._distance_metrics import fidelity
                from ._promotion import promote_state_vector_to_density_matrix

                return bool(jnp.allclose(fidelity(promote_state_vector_to_density_matrix(self), other), 1.0))
            case Unitary() | SuperOp() | Choi() | PauliLiouville() | KrausMap():
                # States and operators are never equal
                return False
            case _:
                return NotImplemented

    def _to_qobj(self) -> "qutip.Qobj | NDArray":
        """Convert to a QuTiP Qobj (or array of Qobjs for ensembles) for interoperability testing."""
        import numpy as np
        import qutip as qt

        dims = [list(self.dims), [1] * len(self.dims)]
        matrix = self.matrix

        if self.ensemble_size == ():
            # Scalar case - return single Qobj
            return qt.Qobj(np.asarray(matrix), dims=dims)

        flat_shape = (-1,) + matrix.shape[-1:]
        flat_vectors = np.asarray(matrix).reshape(flat_shape)
        qobjs = np.asarray(
            [
                qt.Qobj(
                    np.asarray(vec),
                    dims=dims,
                )
                for vec in flat_vectors
            ],
            dtype=object,
        )
        return qobjs.reshape(self.ensemble_size)


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class DensityMatrix(State):
    """Density matrix ρ, shape (*ensemble, d0_out, d1_out, ..., d0_in, d1_in, ...) in tensor form
    or (*ensemble, d, d) in matrix form."""

    @property
    def dims(self) -> Tuple[int, ...]:
        """The dimensions of each qudit, inferred from data shape.

        For DensityMatrix, dims returns just the qudit dimensions (same for in/out).
        """
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        n_qudits = len(qudit_shape) // 2
        return qudit_shape[:n_qudits]

    @property
    def ensemble_size(self) -> Tuple[int, ...]:
        """Returns the size of the ensemble if the state represents an ensemble of states."""
        return self.data.shape[: self.num_ensemble_dims]

    @property
    def matrix(self) -> Array:
        """Returns the matrix representation (*ensemble, d, d) of the density matrix."""
        ensemble_shape = self.data.shape[: self.num_ensemble_dims]
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        n_qudits = len(qudit_shape) // 2
        d_out = reduce(mul, qudit_shape[:n_qudits], 1)
        d_in = reduce(mul, qudit_shape[n_qudits:], 1)
        return self.data.reshape(ensemble_shape + (d_out, d_in))

    @classmethod
    def from_matrix(cls, matrix: Array, dims: Tuple[int, ...], num_ensemble_dims: int = 0) -> "DensityMatrix":
        """Construct from matrix representation.

        :param matrix: Array with shape (*ensemble, d, d) where d = prod(dims)
        :param dims: Tuple of qudit dimensions (d0, d1, ...)
        :param num_ensemble_dims: Number of leading ensemble dimensions
        :return: DensityMatrix with tensor data
        """
        # For density matrices, dims_out = dims_in = dims
        ensemble_shape = matrix.shape[:num_ensemble_dims]
        tensor = matrix.reshape(ensemble_shape + dims + dims)
        return cls(data=tensor, num_ensemble_dims=num_ensemble_dims)

    @property
    def T(self):
        """Transpose of the operator(s)"""
        matrix_t = jnp.swapaxes(self.matrix, -1, -2)
        return DensityMatrix.from_matrix(matrix_t, self.dims, self.num_ensemble_dims)

    @property
    def h(self):
        """Hermitian conjugate of the operator(s)"""
        matrix_h = jnp.conjugate(jnp.swapaxes(self.matrix, -1, -2))
        return DensityMatrix.from_matrix(matrix_h, self.dims, self.num_ensemble_dims)

    def __pow__(self, exponent: float) -> Self:
        """Exponentiation of the density matrix using eigendecomposition (ensemble-compatible)."""
        from ._power import density_matrix_power

        return density_matrix_power(self, exponent)

    def __matmul__(self, other):
        """Left multiply the density matrix by another object."""
        match other:
            case StateVector():  # 𝜌|𝜓> -> |𝜙>
                result = jnp.einsum("...ab,...b->...a", self.matrix, other.matrix)
                num_ensemble = max(self.num_ensemble_dims, other.num_ensemble_dims)
                return StateVector.from_matrix(result, other.dims, num_ensemble)
            case DensityMatrix():  # 𝜌𝜎 -> 𝜏
                result = jnp.einsum("...ab,...bc->...ac", self.matrix, other.matrix)
                num_ensemble = max(self.num_ensemble_dims, other.num_ensemble_dims)
                return DensityMatrix.from_matrix(result, self.dims, num_ensemble)
            case _:
                return NotImplemented

    def __or__(self, other: Any) -> Any:
        """Tensor product of the density matrix with another state."""
        match other:
            case StateVector():
                # ρ ⊗ |ψ⟩ -> DensityMatrix (promote state vector)
                from ._promotion import promote_state_vector_to_density_matrix
                from ._tensor import tensor_density_matrix

                return tensor_density_matrix(self, promote_state_vector_to_density_matrix(other))
            case DensityMatrix():
                # ρ1 ⊗ ρ2 -> DensityMatrix
                from ._tensor import tensor_density_matrix

                return tensor_density_matrix(self, other)
            case Unitary() | SuperOp() | Choi() | PauliLiouville() | KrausMap():
                # State | Operator -> NotImplemented
                return NotImplemented
            case _:
                return NotImplemented

    def __eq__(self, other: Any) -> bool:
        """Check equality using quantum state fidelity."""
        match other:
            case StateVector():
                # Promote other to density matrix and compare
                from ._distance_metrics import fidelity
                from ._promotion import promote_state_vector_to_density_matrix

                return bool(jnp.allclose(fidelity(self, promote_state_vector_to_density_matrix(other)), 1.0))
            case DensityMatrix():
                # Compare two density matrices using fidelity
                from ._distance_metrics import fidelity

                return bool(jnp.allclose(fidelity(self, other), 1.0))
            case Unitary() | SuperOp() | Choi() | PauliLiouville() | KrausMap():
                # States and operators are never equal
                return False
            case _:
                return NotImplemented

    def _to_qobj(self) -> "qutip.Qobj | NDArray":
        """Convert to a QuTiP Qobj (or array of Qobjs for ensembles) for interoperability testing."""
        import numpy as np
        import qutip as qt

        # Density matrices are operators: dims = [input_dims, output_dims] = [dims, dims]
        dims = [list(self.dims), list(self.dims)]  # operator on the same space :contentReference[oaicite:1]{index=1}
        matrix = self.matrix

        if self.ensemble_size == ():
            # Scalar case - return single density-matrix Qobj
            return qt.Qobj(np.asarray(matrix), dims=dims)

        # Ensemble case - matrix has shape ensemble_size + (dim, dim)
        flat_shape = (-1,) + matrix.shape[-2:]
        flat_rhos = np.asarray(matrix).reshape(flat_shape)

        qobjs = np.asarray(
            [qt.Qobj(rho, dims=dims) for rho in flat_rhos],
            dtype=object,
        )
        return qobjs.reshape(self.ensemble_size)


# ---------- operators ----------


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class Unitary(Operator):
    """Unitary operator U, shape (*ensemble, d0_out, d1_out, ..., d0_in, d1_in, ...) in tensor form
    or (*ensemble, d, d) in matrix form."""

    @property
    def dims(self) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """The (output, input) dimensions of each qudit, inferred from data shape."""
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        n_qudits = len(qudit_shape) // 2
        return (qudit_shape[:n_qudits], qudit_shape[n_qudits:])

    def _to_qobj(self) -> "qutip.Qobj | NDArray":
        """Convert to a QuTiP Qobj for interoperability testing."""
        import numpy as np
        import qutip as qt

        dims = [list(self.dims[0]), list(self.dims[1])]
        matrix = self.matrix

        if self.ensemble_size == ():
            return qt.Qobj(
                np.array(matrix),
                dims=dims,
            )

        flat_shape = (-1,) + matrix.shape[-2:]
        flat_unitary = np.asarray(matrix).reshape(flat_shape)
        qobjs = np.asarray(
            [
                qt.Qobj(
                    np.asarray(unitary),
                    dims=dims,
                )
                for unitary in flat_unitary
            ],
            dtype=object,
        )
        return qobjs.reshape(self.ensemble_size)

    def __mul__(self, scalar: complex) -> "Unitary | Kraus":
        """Scalar multiplication of the unitary."""
        if jnp.isclose(jnp.abs(scalar), 1.0):
            return Unitary(self.data * scalar, self.num_ensemble_dims)
        else:
            return Kraus(self.data * scalar, self.num_ensemble_dims)

    def __rmul__(self, scalar: complex) -> "Unitary | Kraus":
        """Scalar multiplication of the unitary."""
        return self * scalar

    def __pow__(self, exponent: float) -> "Unitary":
        """Exponentiation of the unitary using eigendecomposition (ensemble-compatible)."""
        from ._power import power_unitary

        return power_unitary(self, exponent)

    def __matmul__(self, other: Any) -> Any:
        """Matrix multiplication of a quantum object with the unitary."""
        match other:
            case Unitary():
                # U @ U -> Unitary
                from ._compose import compose_unitary

                return compose_unitary(self, other)
            case Choi():
                # J @ U -> Choi (promotion)
                from ._compose import compose_choi
                from ._superoperator_transformations import unitary_to_choi

                return compose_choi(other, unitary_to_choi(self))
            case PauliLiouville():
                # P @ U -> PauliLiouville (promotion)
                from ._compose import compose_pauli_liouville
                from ._superoperator_transformations import unitary_to_pauli_liouville

                return compose_pauli_liouville(other, unitary_to_pauli_liouville(self))
            case SuperOp():
                # S @ U -> SuperOp (promotion)
                from ._compose import compose_superop
                from ._superoperator_transformations import unitary_to_superop

                return compose_superop(other, unitary_to_superop(self))
            case KrausMap():
                # K @ U -> KrausMap (promotion)
                from ._compose import compose_kraus
                from ._superoperator_transformations import unitary_to_kraus

                return compose_kraus(other, unitary_to_kraus(self))
            case StateVector():
                # <psi|U = <phi| -> StateVector (apply unitary to state vector)
                from ._apply import apply_unitary_to_state_vector

                return apply_unitary_to_state_vector(self, other)
            case DensityMatrix():
                # ρU is an Operator product
                from ._apply import apply_superop_to_density_matrix
                from ._superoperator_transformations import unitary_to_superop

                return apply_superop_to_density_matrix(unitary_to_superop(self), other)
            case _:
                return NotImplemented

    def __or__(self, other: Any) -> Any:
        """Tensor product of the unitary with another operator."""
        match other:
            case Unitary():
                # U ⊗ V -> Unitary (tensor product)
                from ._tensor import tensor_unitary

                return tensor_unitary(self, other)
            case SuperOp():
                # U ⊗ S -> SuperOp (promote unitary)
                from ._superoperator_transformations import unitary_to_superop
                from ._tensor import tensor_superop

                return tensor_superop(unitary_to_superop(self), other)
            case Choi():
                # U ⊗ J -> Choi (promote unitary)
                from ._superoperator_transformations import unitary_to_choi
                from ._tensor import tensor_choi

                return tensor_choi(unitary_to_choi(self), other)
            case PauliLiouville():
                # U ⊗ P -> PauliLiouville (promote unitary)
                from ._superoperator_transformations import unitary_to_pauli_liouville
                from ._tensor import tensor_pauli_liouville

                return tensor_pauli_liouville(unitary_to_pauli_liouville(self), other)
            case KrausMap():
                # U ⊗ K -> KrausMap (promote unitary)
                from ._superoperator_transformations import unitary_to_kraus
                from ._tensor import tensor_kraus

                return tensor_kraus(unitary_to_kraus(self), other)
            case StateVector() | DensityMatrix():
                # Operator | State -> NotImplemented
                return NotImplemented
            case _:
                return NotImplemented

    def __eq__(self, other: Any) -> bool:
        """Check equality using unitary entanglement fidelity."""
        match other:
            case Unitary():
                # Compare two unitaries using entanglement fidelity
                from ._distance_metrics import unitary_entanglement_fidelity

                return bool(jnp.allclose(unitary_entanglement_fidelity(self, other), 1.0))
            case SuperOp() | Choi() | PauliLiouville() | KrausMap():
                # Promote self to superoperator and compare using process fidelity
                from ._distance_metrics import process_fidelity
                from ._superoperator_transformations import unitary_to_superop

                return bool(jnp.allclose(process_fidelity(unitary_to_superop(self), other), 1.0))
            case StateVector() | DensityMatrix():
                # States and operators are never equal
                return False
            case _:
                return NotImplemented


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class Kraus(Operator):
    """Kraus superoperator, shape (*ensemble, d0_out, d1_out, ..., d0_in, d1_in, ...) in tensor form
    or (*ensemble, d_out, d_in) in matrix form."""

    @property
    def dims(self) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """The (output, input) dimensions of each qudit, inferred from data shape."""
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        n_qudits = len(qudit_shape) // 2
        return (qudit_shape[:n_qudits], qudit_shape[n_qudits:])

    def _to_qobj(self) -> "qutip.Qobj | NDArray":
        """Convert to a QuTiP Qobj for interoperability testing."""
        import numpy as np
        import qutip as qt

        matrix = self.matrix

        if self.ensemble_size == ():
            return qt.Qobj(
                np.array(matrix),
                dims=[[list(self.dims[0]), list(self.dims[0])], [list(self.dims[1]), list(self.dims[1])]],
            )

        flat_shape = (-1,) + matrix.shape[-2:]
        flat_kraus = np.asarray(matrix).reshape(flat_shape)
        dims = [[list(self.dims[0]), list(self.dims[0])], [list(self.dims[1]), list(self.dims[1])]]
        qobjs = np.asarray(
            [
                qt.Qobj(
                    np.asarray(kraus),
                    dims=dims,
                )
                for kraus in flat_kraus
            ],
            dtype=object,
        )
        return qobjs.reshape(self.ensemble_size)


# ---------- superoperators ----------


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class SuperOp(SuperOperator):
    """SuperOp matrix (also known as Superoperator) S.

    Tensor shape: (*ensemble, d0_out_bra, d1_out_bra, ..., d0_out_ket, d1_out_ket, ...,
                          d0_in_bra, d1_in_bra, ..., d0_in_ket, d1_in_ket, ...)
    Matrix shape: (*ensemble, d_out^2, d_in^2)
    """

    @property
    def dims(self) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """The (output, input) dimensions of each qudit, inferred from data shape.

        Returns the qudit dimensions, not the doubled superoperator dimensions.
        """
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        n_qudits = len(qudit_shape) // 4
        # dims_out is the first n_qudits dimensions (out_bra = out_ket for valid superops)
        # dims_in is the third group of n_qudits dimensions (in_bra = in_ket for valid superops)
        dims_out = qudit_shape[:n_qudits]
        dims_in = qudit_shape[2 * n_qudits : 3 * n_qudits]
        return (dims_out, dims_in)

    def _to_qobj(self) -> "qutip.Qobj | NDArray":
        """Convert to a QuTiP Qobj for interoperability testing."""
        import numpy as np
        import qutip as qt

        matrix = self.matrix

        if self.ensemble_size == ():
            return qt.Qobj(
                np.array(matrix),
                dims=[[list(self.dims[0]), list(self.dims[0])], [list(self.dims[1]), list(self.dims[1])]],
                superrep="super",
            )

        flat_shape = (-1,) + matrix.shape[-2:]
        flat_superop = np.asarray(matrix).reshape(flat_shape)
        dims = [[list(self.dims[0]), list(self.dims[0])], [list(self.dims[1]), list(self.dims[1])]]
        qobjs = np.asarray(
            [
                qt.Qobj(
                    np.asarray(superop),
                    dims=dims,
                    superrep="super",
                )
                for superop in flat_superop
            ],
            dtype=object,
        )
        return qobjs.reshape(self.ensemble_size)

    def __pow__(self, exponent: float) -> Self:
        """Exponentiation of the superoperator using eigendecomposition (ensemble-compatible)."""
        from ._power import power_superop

        return power_superop(self, exponent)

    def __matmul__(self, other: Any) -> Any:
        """Matrix multiplication of the superoperator with another superoperator."""
        match other:
            case SuperOp():
                # S1 @ S2 -> SuperOp (composition)
                from ._compose import compose_superop

                return compose_superop(self, other)
            case Choi():
                # S @ J -> Choi (convert to Choi and compose)
                from ._compose import compose_choi
                from ._superoperator_transformations import superop_to_choi

                return compose_choi(superop_to_choi(self), other)
            case PauliLiouville():
                # S @ P -> PauliLiouville (convert to PauliLiouville and compose)
                from ._compose import compose_pauli_liouville
                from ._superoperator_transformations import superop_to_pauli_liouville

                return compose_pauli_liouville(superop_to_pauli_liouville(self), other)
            case KrausMap():
                # S @ K -> KrausMap (convert to KrausMap and compose)
                from ._compose import compose_kraus
                from ._superoperator_transformations import superop_to_kraus

                return compose_kraus(superop_to_kraus(self), other)
            case Unitary():
                # S @ U -> SuperOp (promotion)
                from ._compose import compose_superop
                from ._superoperator_transformations import unitary_to_superop

                return compose_superop(self, unitary_to_superop(other))
            case StateVector():
                # S @ |ψ⟩ -> DensityMatrix (promotion)
                from ._apply import apply_superop_to_density_matrix
                from ._promotion import promote_state_vector_to_density_matrix

                return apply_superop_to_density_matrix(self, promote_state_vector_to_density_matrix(other))
            case DensityMatrix():
                # S @ ρ -> DensityMatrix (apply channel to density matrix)
                from ._apply import apply_superop_to_density_matrix

                return apply_superop_to_density_matrix(self, other)
            case _:
                return NotImplemented

    def __or__(self, other: Any) -> Any:
        """Tensor product of the superoperator with another operator."""
        match other:
            case SuperOp():
                # S1 ⊗ S2 -> SuperOp
                from ._tensor import tensor_superop

                return tensor_superop(self, other)
            case Choi():
                # S ⊗ J -> Choi (convert to Choi)
                from ._superoperator_transformations import superop_to_choi
                from ._tensor import tensor_choi

                return tensor_choi(superop_to_choi(self), other)
            case PauliLiouville():
                # S ⊗ P -> PauliLiouville (convert to PauliLiouville)
                from ._superoperator_transformations import superop_to_pauli_liouville
                from ._tensor import tensor_pauli_liouville

                return tensor_pauli_liouville(superop_to_pauli_liouville(self), other)
            case KrausMap():
                # S ⊗ K -> KrausMap (convert to KrausMap)
                from ._superoperator_transformations import superop_to_kraus
                from ._tensor import tensor_kraus

                return tensor_kraus(superop_to_kraus(self), other)
            case Unitary():
                # S ⊗ U -> SuperOp (promote unitary)
                from ._superoperator_transformations import unitary_to_superop
                from ._tensor import tensor_superop

                return tensor_superop(self, unitary_to_superop(other))
            case StateVector() | DensityMatrix():
                # Operator | State -> NotImplemented
                return NotImplemented
            case _:
                return NotImplemented

    def __eq__(self, other: Any) -> bool:
        """Check equality using process fidelity."""
        match other:
            case SuperOp():
                # Compare two superoperators using process fidelity
                from ._distance_metrics import process_fidelity

                return bool(jnp.allclose(process_fidelity(self, other), 1.0))
            case Choi() | PauliLiouville() | KrausMap():
                # Convert other to SuperOp and compare
                from ._distance_metrics import process_fidelity

                return bool(jnp.allclose(process_fidelity(self, other), 1.0))
            case Unitary():
                # Promote Unitary to SuperOp and compare
                from ._distance_metrics import process_fidelity
                from ._superoperator_transformations import unitary_to_superop

                return bool(jnp.allclose(process_fidelity(self, unitary_to_superop(other)), 1.0))
            case StateVector() | DensityMatrix():
                # States and operators are never equal
                return False
            case _:
                return NotImplemented


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class KrausMap(SuperOperator):
    """Kraus channel.

    Tensor shape: (*ensemble, n_kraus, d0_out, d1_out, ..., d0_in, d1_in, ...)
    Matrix shape: (*ensemble, n_kraus, d_out, d_in)
    """

    @property
    def dims(self) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """The (output, input) dimensions of each qudit, inferred from data shape."""
        # After ensemble dims, first is n_kraus, then qudit dims
        qudit_shape = self.data.shape[self.num_ensemble_dims + 1 :]
        n_qudits = len(qudit_shape) // 2
        return (qudit_shape[:n_qudits], qudit_shape[n_qudits:])

    @property
    def ensemble_size(self) -> Tuple[int, ...]:
        """Returns the size of the ensemble if the operator represents an ensemble of operators."""
        return self.data.shape[: self.num_ensemble_dims]

    @property
    def matrix(self) -> Array:
        """Returns the matrix representation (*ensemble, n_kraus, d_out, d_in) of the Kraus map."""
        ensemble_shape = self.data.shape[: self.num_ensemble_dims]
        n_kraus = self.data.shape[self.num_ensemble_dims]
        qudit_shape = self.data.shape[self.num_ensemble_dims + 1 :]
        n_qudits = len(qudit_shape) // 2
        d_out = reduce(mul, qudit_shape[:n_qudits], 1)
        d_in = reduce(mul, qudit_shape[n_qudits:], 1)
        return self.data.reshape(ensemble_shape + (n_kraus, d_out, d_in))

    @classmethod
    def from_matrix(
        cls, matrix: Array, dims: Tuple[Tuple[int, ...], Tuple[int, ...]], num_ensemble_dims: int = 0
    ) -> "KrausMap":
        """Construct from matrix representation.

        :param matrix: Array with shape (*ensemble, n_kraus, d_out, d_in)
        :param dims: Tuple of (dims_out, dims_in) where each is a tuple of qudit dimensions
        :param num_ensemble_dims: Number of leading ensemble dimensions
        :return: KrausMap with tensor data
        """
        ensemble_shape = matrix.shape[:num_ensemble_dims]
        n_kraus = matrix.shape[num_ensemble_dims]
        tensor = matrix.reshape(ensemble_shape + (n_kraus,) + dims[0] + dims[1])
        return cls(data=tensor, num_ensemble_dims=num_ensemble_dims)

    def _to_qobj(self) -> list["qutip.Qobj"]:
        """Convert to a QuTiP Qobj for interoperability testing.

        Returns a list of QuTiP Qobjs, one for each Kraus operator.
        """
        import numpy as np
        import qutip as qt

        matrix = self.matrix
        # KrausMap is always an ensemble of Kraus operators, shape (K, d_out, d_in)
        return [qt.Qobj(np.array(k), dims=[[list(self.dims[0])], [list(self.dims[1])]]) for k in matrix]

    def __pow__(self, exponent: float) -> Self:
        """Exponentiation of the Kraus channel using eigendecomposition (ensemble-compatible)."""
        from ._power import power_kraus

        return power_kraus(self, exponent)

    def __matmul__(self, other: Any) -> Any:
        """Matrix multiplication of the Kraus channel with another superoperator."""
        match other:
            case KrausMap():
                # K1 @ K2 -> KrausMap (composition)
                from ._compose import compose_kraus

                return compose_kraus(self, other)
            case SuperOp():
                # K @ S -> SuperOp (convert to SuperOp and compose)
                from ._compose import compose_superop
                from ._superoperator_transformations import kraus_to_superop

                return compose_superop(kraus_to_superop(self), other)
            case Choi():
                # K @ J -> Choi (convert to Choi and compose)
                from ._compose import compose_choi
                from ._superoperator_transformations import kraus_to_choi

                return compose_choi(kraus_to_choi(self), other)
            case PauliLiouville():
                # K @ P -> PauliLiouville (convert to PauliLiouville and compose)
                from ._compose import compose_pauli_liouville
                from ._superoperator_transformations import kraus_to_pauli_liouville

                return compose_pauli_liouville(kraus_to_pauli_liouville(self), other)
            case Unitary():
                # K @ U -> KrausMap (promotion)
                from ._compose import compose_kraus
                from ._superoperator_transformations import unitary_to_kraus

                return compose_kraus(self, unitary_to_kraus(other))
            case StateVector():
                # K @ |ψ⟩ -> DensityMatrix (promotion)
                from ._apply import apply_kraus_to_density_matrix
                from ._promotion import promote_state_vector_to_density_matrix

                return apply_kraus_to_density_matrix(self, promote_state_vector_to_density_matrix(other))
            case DensityMatrix():
                # K @ ρ -> DensityMatrix (apply channel to density matrix)
                from ._apply import apply_kraus_to_density_matrix

                return apply_kraus_to_density_matrix(self, other)
            case _:
                return NotImplemented

    def __or__(self, other: Any) -> Any:
        """Tensor product of the Kraus channel with another quantum object."""
        match other:
            case KrausMap():
                # K1 | K2 -> KrausMap
                from ._tensor import tensor_kraus

                return tensor_kraus(self, other)
            case SuperOp():
                # K | S -> SuperOp (convert to SuperOp and tensor)
                from ._superoperator_transformations import kraus_to_superop
                from ._tensor import tensor_superop

                return tensor_superop(kraus_to_superop(self), other)
            case Choi():
                # K | J -> Choi (convert to Choi and tensor)
                from ._superoperator_transformations import kraus_to_choi
                from ._tensor import tensor_choi

                return tensor_choi(kraus_to_choi(self), other)
            case PauliLiouville():
                # K | P -> PauliLiouville (convert to PauliLiouville and tensor)
                from ._superoperator_transformations import kraus_to_pauli_liouville
                from ._tensor import tensor_pauli_liouville

                return tensor_pauli_liouville(kraus_to_pauli_liouville(self), other)
            case Unitary():
                # K | U -> KrausMap (promote Unitary to KrausMap and tensor)
                from ._superoperator_transformations import unitary_to_kraus
                from ._tensor import tensor_kraus

                return tensor_kraus(self, unitary_to_kraus(other))
            case StateVector() | DensityMatrix():
                # K | |ψ⟩ or K | ρ -> NotImplemented (operator | state)
                return NotImplemented
            case _:
                return NotImplemented

    def __eq__(self, other: Any) -> bool:
        """Check equality using process fidelity."""
        match other:
            case KrausMap():
                # Compare two KrausMaps using process fidelity
                from ._distance_metrics import process_fidelity

                return bool(jnp.allclose(process_fidelity(self, other), 1.0))
            case SuperOp() | Choi() | PauliLiouville():
                # Compare using process fidelity (handles conversions internally)
                from ._distance_metrics import process_fidelity

                return bool(jnp.allclose(process_fidelity(self, other), 1.0))
            case Unitary():
                # Promote Unitary to KrausMap and compare
                from ._distance_metrics import process_fidelity
                from ._superoperator_transformations import unitary_to_kraus

                return bool(jnp.allclose(process_fidelity(self, unitary_to_kraus(other)), 1.0))
            case StateVector() | DensityMatrix():
                # States and operators are never equal
                return False
            case _:
                return NotImplemented


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class Choi(SuperOperator):
    """Choi matrix C.

    Tensor shape: (*ensemble, d0_out_bra, d1_out_bra, ..., d0_out_ket, d1_out_ket, ...,
                          d0_in_bra, d1_in_bra, ..., d0_in_ket, d1_in_ket, ...)
    Matrix shape: (*ensemble, d_out^2, d_in^2)
    """

    @property
    def dims(self) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """The (output, input) dimensions of each qudit, inferred from data shape."""
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        n_qudits = len(qudit_shape) // 4
        dims_out = qudit_shape[:n_qudits]
        dims_in = qudit_shape[2 * n_qudits : 3 * n_qudits]
        return (dims_out, dims_in)

    def _to_qobj(self) -> "qutip.Qobj | NDArray":
        """Convert to a QuTiP Qobj for interoperability testing."""
        import numpy as np
        import qutip as qt

        matrix = self.matrix

        if self.ensemble_size == ():
            return qt.Qobj(
                np.array(matrix),
                dims=[[list(self.dims[0]), list(self.dims[0])], [list(self.dims[1]), list(self.dims[1])]],
                superrep="choi",
            )

        flat_shape = (-1,) + matrix.shape[-2:]
        flat_choi = np.asarray(matrix).reshape(flat_shape)
        dims = [[list(self.dims[0]), list(self.dims[0])], [list(self.dims[1]), list(self.dims[1])]]
        qobjs = np.asarray(
            [
                qt.Qobj(
                    np.asarray(choi),
                    dims=dims,
                    superrep="choi",
                )
                for choi in flat_choi
            ],
            dtype=object,
        )
        return qobjs.reshape(self.ensemble_size)

    def __pow__(self, exponent: float) -> Self:
        """Exponentiation of the Choi matrix using eigendecomposition (ensemble-compatible)."""
        from ._power import power_choi

        return power_choi(self, exponent)

    def __matmul__(self, other: Any) -> Any:
        """Matrix multiplication of the Choi with another Superoperator."""
        match other:
            case Choi():
                # J1 @ J2 -> Choi (composition)
                from ._compose import compose_choi

                return compose_choi(self, other)
            case SuperOp():
                # J @ S -> SuperOp (convert to SuperOp and compose)
                from ._compose import compose_superop
                from ._superoperator_transformations import choi_to_superop

                return compose_superop(choi_to_superop(self), other)
            case PauliLiouville():
                # J @ P -> PauliLiouville (convert to PauliLiouville and compose)
                from ._compose import compose_pauli_liouville
                from ._superoperator_transformations import choi_to_pauli_liouville

                return compose_pauli_liouville(choi_to_pauli_liouville(self), other)
            case KrausMap():
                # J @ K -> KrausMap (convert to KrausMap and compose)
                from ._compose import compose_kraus
                from ._superoperator_transformations import choi_to_kraus

                return compose_kraus(choi_to_kraus(self), other)
            case Unitary():
                # J @ U -> Choi (promotion)
                from ._compose import compose_choi
                from ._superoperator_transformations import unitary_to_choi

                return compose_choi(self, unitary_to_choi(other))
            case StateVector():
                # J @ |ψ⟩ -> DensityMatrix (apply channel to state)
                from ._apply import apply_choi_to_density_matrix
                from ._promotion import promote_state_vector_to_density_matrix

                return apply_choi_to_density_matrix(self, promote_state_vector_to_density_matrix(other))
            case DensityMatrix():
                # J @ ρ -> DensityMatrix (apply channel to density matrix)
                from ._apply import apply_choi_to_density_matrix

                return apply_choi_to_density_matrix(self, other)
            case _:
                return NotImplemented

    def __or__(self, other: Any) -> Any:
        """Tensor product of the Choi matrix with another quantum object."""
        match other:
            case Choi():
                # J1 | J2 -> Choi
                from ._tensor import tensor_choi

                return tensor_choi(self, other)
            case SuperOp():
                # J | S -> SuperOp (convert to SuperOp and tensor)
                from ._superoperator_transformations import choi_to_superop
                from ._tensor import tensor_superop

                return tensor_superop(choi_to_superop(self), other)
            case PauliLiouville():
                # J | P -> PauliLiouville (convert to PauliLiouville and tensor)
                from ._superoperator_transformations import choi_to_pauli_liouville
                from ._tensor import tensor_pauli_liouville

                return tensor_pauli_liouville(choi_to_pauli_liouville(self), other)
            case KrausMap():
                # J | K -> KrausMap (convert to KrausMap and tensor)
                from ._superoperator_transformations import choi_to_kraus
                from ._tensor import tensor_kraus

                return tensor_kraus(choi_to_kraus(self), other)
            case Unitary():
                # J | U -> Choi (promote Unitary to Choi and tensor)
                from ._superoperator_transformations import unitary_to_choi
                from ._tensor import tensor_choi

                return tensor_choi(self, unitary_to_choi(other))
            case StateVector() | DensityMatrix():
                # J | |ψ⟩ or J | ρ -> NotImplemented (operator | state)
                return NotImplemented
            case _:
                return NotImplemented

    def __eq__(self, other: Any) -> bool:
        """Check equality using process fidelity."""
        match other:
            case Choi():
                # Compare two Choi matrices using process fidelity
                from ._distance_metrics import process_fidelity

                return bool(jnp.allclose(process_fidelity(self, other), 1.0))
            case SuperOp() | PauliLiouville() | KrausMap():
                # Compare using process fidelity (handles conversions internally)
                from ._distance_metrics import process_fidelity

                return bool(jnp.allclose(process_fidelity(self, other), 1.0))
            case Unitary():
                # Promote Unitary to Choi and compare
                from ._distance_metrics import process_fidelity
                from ._superoperator_transformations import unitary_to_choi

                return bool(jnp.allclose(process_fidelity(self, unitary_to_choi(other)), 1.0))
            case StateVector() | DensityMatrix():
                # States and operators are never equal
                return False
            case _:
                return NotImplemented


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class Chi(SuperOperator):
    """Chi matrix Χ.

    Tensor shape: (*ensemble, d0_out_bra, d1_out_bra, ..., d0_out_ket, d1_out_ket, ...,
                          d0_in_bra, d1_in_bra, ..., d0_in_ket, d1_in_ket, ...)
    Matrix shape: (*ensemble, d_out^2, d_in^2)
    """

    @property
    def dims(self) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """The (output, input) dimensions of each qudit, inferred from data shape."""
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        n_qudits = len(qudit_shape) // 4
        dims_out = qudit_shape[:n_qudits]
        dims_in = qudit_shape[2 * n_qudits : 3 * n_qudits]
        return (dims_out, dims_in)

    def _to_qobj(self) -> "qutip.Qobj | list[qutip.Qobj]":
        """Convert to a QuTiP Qobj for interoperability testing."""
        import numpy as np
        import qutip as qt

        matrix = self.matrix

        if self.ensemble_size != ():
            # Batched Chi matrices - return list
            chi_qobjs = []
            for c in matrix:
                base_qobj = qt.Qobj(
                    np.array(c),
                    dims=[[list(self.dims[0]), list(self.dims[0])], [list(self.dims[1]), list(self.dims[1])]],
                )
                chi_qobjs.append(qt.to_chi(base_qobj))
            return chi_qobjs
        # Single Chi matrix
        base_qobj = qt.Qobj(
            np.array(matrix),
            dims=[[list(self.dims[0]), list(self.dims[0])], [list(self.dims[1]), list(self.dims[1])]],
        )
        return qt.to_chi(base_qobj)

    def __pow__(self, exponent: float) -> Self:
        return NotImplemented

    def __matmul__(self, other: Any) -> Any:
        """Matrix multiplication of the Chi matrix with another superoperator."""
        match other:
            case Chi():
                # Composition of Chi matrices is not directly supported.
                # One would need to convert to another representation (e.g., SuperOp),
                # compose, and then convert back.
                # No chi transformations are implemented
                raise NotImplementedError("Composition of Chi matrices is not implemented.")
            case SuperOp():
                # χ @ S -> SuperOp (convert Chi to SuperOp and compose)
                raise NotImplementedError("Chi to SuperOp conversion not implemented.")
            case Choi():
                # χ @ J -> Choi (convert Chi to Choi and compose)
                raise NotImplementedError("Chi to Choi conversion not implemented.")
            case PauliLiouville():
                # χ @ P -> PauliLiouville (convert Chi to PauliLiouville and compose)
                raise NotImplementedError("Chi to PauliLiouville conversion not implemented.")
            case KrausMap():
                # χ @ K -> KrausMap (convert Chi to KrausMap and compose)
                raise NotImplementedError("Chi to KrausMap conversion not implemented.")
            case Unitary():
                # χ @ U -> Chi (promotion)
                raise NotImplementedError("Unitary to Chi conversion not implemented.")
            case StateVector():
                # χ @ |ψ⟩ -> DensityMatrix (convert to applicable representation)
                raise NotImplementedError("Applying Chi to StateVector is not implemented.")
            case DensityMatrix():
                # χ @ ρ -> DensityMatrix (convert to applicable representation)
                raise NotImplementedError("Applying Chi to DensityMatrix is not implemented.")
            case _:
                return NotImplemented

    def __or__(self, other: Any) -> Any:
        """Tensor product of the Chi matrix with another quantum object."""
        # Since Chi transformations are not implemented, we can only handle Chi | Chi
        match other:
            case Chi():
                # χ1 | χ2 -> Chi
                raise NotImplementedError("Tensor product of Chi matrices is not implemented.")
            case SuperOp() | Choi() | PauliLiouville() | KrausMap() | Unitary():
                # χ | S -> would require Chi transformations (not implemented)
                raise NotImplementedError(f"Tensor product not implemented between Chi and {type(other).__name__}.")
            case StateVector() | DensityMatrix():
                # χ | |ψ⟩ or χ | ρ -> NotImplemented (operator | state)
                return NotImplemented
            case _:
                return NotImplemented

    def __eq__(self, other: Any) -> bool:
        """Check equality for Chi matrices."""
        # Since Chi transformations are not implemented, we can only compare Chi with Chi
        match other:
            case Chi():
                # Direct comparison not possible without Chi transformations
                raise NotImplementedError("Equality comparison for Chi matrices is not implemented.")
            case SuperOp() | Choi() | PauliLiouville() | KrausMap() | Unitary():
                # Would require Chi transformations (not implemented)
                raise NotImplementedError(
                    "Equality comparison between Chi and other superoperators is not implemented."
                )
            case StateVector() | DensityMatrix():
                # States and operators are never equal
                return False
            case _:
                return NotImplemented


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class PauliLiouville(SuperOperator):
    """Pauli-Liouville matrix P.

    Tensor shape: (*ensemble, d0_out_bra, d1_out_bra, ..., d0_out_ket, d1_out_ket, ...,
                          d0_in_bra, d1_in_bra, ..., d0_in_ket, d1_in_ket, ...)
    Matrix shape: (*ensemble, d_out^2, d_in^2)
    """

    @property
    def dims(self) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        """The (output, input) dimensions of each qudit, inferred from data shape."""
        qudit_shape = self.data.shape[self.num_ensemble_dims :]
        n_qudits = len(qudit_shape) // 4
        dims_out = qudit_shape[:n_qudits]
        dims_in = qudit_shape[2 * n_qudits : 3 * n_qudits]
        return (dims_out, dims_in)

    def _to_qobj(self):
        """Convert to a QuTiP Qobj for interoperability testing.

        Note: QuTiP doesn't have native Pauli-Liouville representation,
        so this will look like a SuperOp Qobj, but it will be the Pauli-Liouville matrix.
        """
        raise NotImplementedError("Conversion to QuTiP Qobj not implemented for PauliLiouville.")

    def __pow__(self, exponent: float) -> Self:
        """Exponentiation of the Pauli-Liouville matrix using eigendecomposition (ensemble-compatible)."""
        from ._power import power_pauli_liouville

        return power_pauli_liouville(self, exponent)

    def __matmul__(self, other: Any) -> Any:
        """Matrix multiplication of the PauliLiouville with another PauliLiouville."""
        match other:
            case PauliLiouville():
                # P1 @ P2 -> PauliLiouville (composition)
                from ._compose import compose_pauli_liouville

                return compose_pauli_liouville(self, other)
            case SuperOp():
                # P @ S -> SuperOp (convert to SuperOp and compose)
                from ._compose import compose_superop
                from ._superoperator_transformations import pauli_liouville_to_superop

                return compose_superop(pauli_liouville_to_superop(self), other)
            case Choi():
                # P @ J -> Choi (convert to Choi and compose)
                from ._compose import compose_choi
                from ._superoperator_transformations import pauli_liouville_to_choi

                return compose_choi(pauli_liouville_to_choi(self), other)
            case KrausMap():
                # P @ K -> KrausMap (convert to KrausMap and compose)
                from ._compose import compose_kraus
                from ._superoperator_transformations import pauli_liouville_to_kraus

                return compose_kraus(pauli_liouville_to_kraus(self), other)
            case Unitary():
                # P @ U -> PauliLiouville (promotion)
                from ._compose import compose_pauli_liouville
                from ._superoperator_transformations import unitary_to_pauli_liouville

                return compose_pauli_liouville(self, unitary_to_pauli_liouville(other))
            case StateVector():
                # P @ |ψ⟩ -> DensityMatrix (apply channel to state)
                from ._apply import apply_pauli_liouville_to_density_matrix
                from ._promotion import promote_state_vector_to_density_matrix

                return apply_pauli_liouville_to_density_matrix(self, promote_state_vector_to_density_matrix(other))
            case DensityMatrix():
                # P @ ρ -> DensityMatrix (apply channel to density matrix)
                from ._apply import apply_pauli_liouville_to_density_matrix

                return apply_pauli_liouville_to_density_matrix(self, other)
            case _:
                return NotImplemented

    def __or__(self, other: Any) -> Any:
        """Tensor product of the Pauli-Liouville matrix with another quantum object."""
        match other:
            case PauliLiouville():
                # P1 | P2 -> PauliLiouville
                from ._tensor import tensor_pauli_liouville

                return tensor_pauli_liouville(self, other)
            case SuperOp():
                # P | S -> PauliLiouville (convert to PauliLiouville and tensor)
                from ._superoperator_transformations import pauli_liouville_to_superop
                from ._tensor import tensor_superop

                return tensor_superop(pauli_liouville_to_superop(self), other)
            case Choi():
                # P | J -> PauliLiouville (convert to PauliLiouville and tensor)
                from ._superoperator_transformations import pauli_liouville_to_choi
                from ._tensor import tensor_choi

                return tensor_choi(pauli_liouville_to_choi(self), other)
            case KrausMap():
                # P | K -> PauliLiouville (convert to PauliLiouville and tensor)
                from ._superoperator_transformations import pauli_liouville_to_kraus
                from ._tensor import tensor_kraus

                return tensor_kraus(pauli_liouville_to_kraus(self), other)
            case Unitary():
                # P | U -> PauliLiouville (promote Unitary to PauliLiouville and tensor)
                from ._superoperator_transformations import unitary_to_pauli_liouville
                from ._tensor import tensor_pauli_liouville

                return tensor_pauli_liouville(self, unitary_to_pauli_liouville(other))
            case StateVector() | DensityMatrix():
                # P | |ψ⟩ or P | ρ -> NotImplemented (operator | state)
                return NotImplemented
            case _:
                return NotImplemented

    def __eq__(self, other: Any) -> bool:
        """Check equality using process fidelity."""
        match other:
            case PauliLiouville():
                # Compare two PauliLiouville matrices using process fidelity
                from ._distance_metrics import process_fidelity

                return bool(jnp.allclose(process_fidelity(self, other), 1.0))
            case SuperOp() | Choi() | KrausMap():
                # Compare using process fidelity (handles conversions internally)
                from ._distance_metrics import process_fidelity

                return bool(jnp.allclose(process_fidelity(self, other), 1.0))
            case Unitary():
                # Promote Unitary to PauliLiouville and compare
                from ._distance_metrics import process_fidelity
                from ._superoperator_transformations import unitary_to_pauli_liouville

                return bool(jnp.allclose(process_fidelity(self, unitary_to_pauli_liouville(other)), 1.0))
            case StateVector() | DensityMatrix():
                # States and operators are never equal
                return False
            case _:
                return NotImplemented



# Copyright 2026 Rigetti & Co, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
import jax.numpy as jnp

from ._quantum_objects import DensityMatrix, StateVector


@jax.jit(static_argnames=("n_qubits",))
def zero_state_vector(n_qubits: int) -> StateVector:
    """
    Construct a vector corresponding to ``|0>``.

    :param n_qubits: The number of qubits.
    :return: The state vector ``|000...0>`` for `n_qubits`.
    """
    state_matrix = jnp.zeros((2**n_qubits), complex)
    state_matrix = state_matrix.at[0].set(complex(1.0, 0))
    return StateVector.from_matrix(state_matrix, (2,) * n_qubits, 0)


@jax.jit(static_argnames=("n_qubits",))
def zero_state_matrix(n_qubits: int) -> DensityMatrix:
    """
    Construct a matrix corresponding to ``|0><0|``.

    :param n_qubits: The number of qubits.
    :return: The state matrix ``|000...0><000...0|`` for `n_qubits`.
    """
    state_matrix = jnp.zeros((2**n_qubits, 2**n_qubits), complex)
    state_matrix = state_matrix.at[0, 0].set(complex(1.0, 0))
    return DensityMatrix.from_matrix(state_matrix, (2,) * n_qubits, 0)


@jax.jit(static_argnames=("n_qubits",))
def mixed_state_matrix(n_qubits: int) -> DensityMatrix:
    """
    Construct a matrix corresponding to the maximally mixed state.

    :param n_qubits: The number of qubits.
    :return: The state matrix  ``I / d`` where ``d = 2**n_qubits``.
    """
    d = 2**n_qubits
    state_matrix = jnp.eye(d, dtype=complex) / d
    return DensityMatrix.from_matrix(state_matrix, (2,) * n_qubits, 0)


@jax.jit
def tensor_state_vectors(state_a: StateVector, state_b: StateVector) -> StateVector:
    """
    Compute the tensor product of two state vectors.

    :param state_a: The first state vector.
    :param state_b: The second state vector.
    :return: The tensor product state vector.
    """
    new_data = jnp.kron(state_a.matrix, state_b.matrix)
    new_dims = state_a.dims + state_b.dims
    return StateVector.from_matrix(new_data, new_dims, 0)


@jax.jit
def tensor_density_matrices(state_a: DensityMatrix, state_b: DensityMatrix) -> DensityMatrix:
    """
    Compute the tensor product of two density matrices.

    :param state_a: The first density matrix.
    :param state_b: The second density matrix.
    :return: The tensor product density matrix.
    """
    new_data = jnp.kron(state_a.matrix, state_b.matrix)
    new_dims = state_a.dims + state_b.dims
    return DensityMatrix.from_matrix(new_data, new_dims, 0)



# Copyright 2026 Rigetti & Co, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""JAX-based quantum operator transformations."""

from ._apply import (
    apply_choi_to_density_matrix,
    apply_kraus_to_density_matrix,
    apply_kraus_to_state_vector,
    apply_pauli_liouville_to_density_matrix,
    apply_superop_to_density_matrix,
    apply_unitary_to_state_vector,
    compute_choi_observables_from_states,
    compute_kraus_observables_from_states,
    compute_pauli_liouville_observables_from_states,
    compute_superop_observables_from_states,
    partial_trace,
)
from ._common_channels import (
    depolarizing_channel_superoperator,
    integrated_thermal_superoperator,
    thermal_relaxation_choi,
    bit_flip_operators,
    phase_flip_operators,
    bitphase_flip_operators,
    dephasing_operators,
    depolarizing_operators,
    relaxation_operators,
    KRAUS_OPS,
)
from ._compose import (
    compose_choi,
    compose_kraus,
    compose_pauli_liouville,
    compose_superop,
    compose_unitary,
)
from ._distance_metrics import (
    average_fidelity_to_depolarizing_constant,
    average_fidelity_to_process_fidelity,
    depolarizing_constant_to_average_fidelity,
    depolarizing_constant_to_process_fidelity,
    fidelity,
    process_fidelity,
    process_fidelity_to_average_fidelity,
    process_fidelity_to_depolarizing_constant,
    unitarity_to_stochastic_infidelity,
    unitary_entanglement_fidelity,
)
from ._power import power_choi, power_kraus, power_pauli_liouville, power_superop, power_unitary
from ._promotion import promote_state_vector_to_density_matrix
from ._quantum_objects import (
    Choi,
    DensityMatrix,
    Kraus,
    KrausMap,
    Operator,
    PauliLiouville,
    State,
    StateVector,
    SuperOp,
    SuperOperator,
    Unitary,
)
from ._random import (
    ginibre_matrix_complex,
    random_choi_BCSZ,
    random_density_matrix,
    random_state_vector,
    random_unitary,
)
from ._state import (
    mixed_state_matrix,
    tensor_density_matrices,
    tensor_state_vectors,
    zero_state_matrix,
    zero_state_vector,
)
from ._superoperator_transformations import (
    choi_to_kraus,
    choi_to_pauli_liouville,
    choi_to_superop,
    kraus_to_choi,
    kraus_to_pauli_liouville,
    kraus_to_superop,
    pauli_liouville_to_choi,
    pauli_liouville_to_kraus,
    pauli_liouville_to_superop,
    superop_to_choi,
    superop_to_kraus,
    superop_to_pauli_liouville,
    to_choi,
    to_kraus,
    to_pauli_liouville,
    to_superop,
    unitary_to_choi,
    unitary_to_pauli_liouville,
    unitary_to_superop,
)
from ._tensor import (
    tensor_choi,
    tensor_density_matrix,
    tensor_kraus,
    tensor_pauli_liouville,
    tensor_state_vector,
    tensor_superop,
    tensor_unitary,
)
from ._validation import (
    is_completely_positive,
    is_cptp,
    is_identity_matrix,
    is_one_design,
    is_trace_preserving,
    is_two_design,
    is_unitary,
)

# Import gates, states and ensembles as submodules
from . import gates
from . import states
from . import ensembles

__all__ = [
    # Apply superoperator functions
    "compute_choi_observables_from_states",
    "compute_kraus_observables_from_states",
    "compute_pauli_liouville_observables_from_states",
    "compute_superop_observables_from_states",
    "partial_trace",
    "apply_choi_to_density_matrix",
    "apply_kraus_to_density_matrix",
    "apply_pauli_liouville_to_density_matrix",
    "apply_superop_to_density_matrix",
    "apply_unitary_to_state_vector",
    "apply_kraus_to_state_vector",
    # Common channels
    "depolarizing_channel_superoperator",
    "integrated_thermal_superoperator",
    "thermal_relaxation_choi",
    "bit_flip_operators",
    "phase_flip_operators",
    "bitphase_flip_operators",
    "dephasing_operators",
    "depolarizing_operators",
    "relaxation_operators",
    "KRAUS_OPS",
    # Compose quantum objects
    "compose_kraus",
    "compose_choi",
    "compose_pauli_liouville",
    "compose_superop",
    "compose_unitary",
    # Tensor quantum objects
    "tensor_kraus",
    "tensor_choi",
    "tensor_pauli_liouville",
    "tensor_superop",
    "tensor_unitary",
    "tensor_state_vector",
    "tensor_density_matrix",
    # Superoperator transformations
    "choi_to_kraus",
    "choi_to_pauli_liouville",
    "choi_to_superop",
    "kraus_to_choi",
    "kraus_to_pauli_liouville",
    "kraus_to_superop",
    "pauli_liouville_to_choi",
    "pauli_liouville_to_kraus",
    "pauli_liouville_to_superop",
    "superop_to_choi",
    "superop_to_kraus",
    "superop_to_pauli_liouville",
    "to_choi",
    "to_pauli_liouville",
    "to_superop",
    "to_kraus",
    "unitary_to_choi",
    "unitary_to_pauli_liouville",
    "unitary_to_superop",
    # Distance metrics
    "fidelity",
    "process_fidelity",
    "to_choi",
    "unitary_entanglement_fidelity",
    "depolarizing_constant_to_average_fidelity",
    "average_fidelity_to_depolarizing_constant",
    "depolarizing_constant_to_process_fidelity",
    "process_fidelity_to_depolarizing_constant",
    "average_fidelity_to_process_fidelity",
    "process_fidelity_to_average_fidelity",
    "unitarity_to_stochastic_infidelity",
    # Types
    "Choi",
    "Kraus",
    "SuperOp",
    "KrausMap",
    "PauliLiouville",
    "SuperOperator",
    "Unitary",
    "DensityMatrix",
    "StateVector",
    "State",
    "Operator",
    # State functions
    "zero_state_matrix",
    "zero_state_vector",
    "mixed_state_matrix",
    "tensor_density_matrices",
    "tensor_state_vectors",
    # Random functions
    "ginibre_matrix_complex",
    "random_density_matrix",
    "random_unitary",
    "random_choi_BCSZ",
    "random_state_vector",
    # Validation functions
    "is_unitary",
    "is_one_design",
    "is_two_design",
    "is_identity_matrix",
    "is_cptp",
    "is_completely_positive",
    "is_trace_preserving",
    # Promotion functions
    "promote_state_vector_to_density_matrix",
    # Power functions
    "power_choi",
    "power_kraus",
    "power_pauli_liouville",
    "power_superop",
    "power_unitary",
    # Submodules
    "gates",
    "states",
    "ensembles",
]



# Copyright 2021-2023 Rigetti & Co, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""
operator_tools.common_channels module
-----------------------------------------
A module containing common channels.
"""

from functools import partial

import jax
import jax.numpy as jnp
from jax import Array, jit

from ._compose import compose_superop
from ._quantum_objects import Choi, SuperOp, Unitary, Kraus
from ._superoperator_transformations import (
    choi_to_superop,
    unitary_to_superop,
)
from ._tensor import tensor_choi
from .gates import X, Y, Z, I


@partial(jit, static_argnums=())
def _thermal_relaxation_choi_1q(t1: float, tphi: float, duration: float) -> Choi:
    """
    Construct the Choi matrix for a single qubit thermal relaxation channel.

    :param t1: The T1 time.
    :param tphi: The pure dephasing time.
    :param duration: The duration.
    :return: The 4x4 Choi matrix for the single-qubit thermal relaxation channel.
    """
    e1 = jnp.exp(-duration / t1)
    # 1/tphi = 1/t2 - 1/2t1
    t2 = 1 / (1 / tphi + 1 / (2 * t1))
    e2 = jnp.exp(-duration / t2)

    return Choi.from_matrix(
        jnp.array(
            [
                [1, 0, 0, e2],
                [0, 0, 0, 0],
                [0, 0, 1 - e1, 0],
                [e2, 0, 0, e1],
            ],
            dtype=jnp.complex64,
        ),
        ((2,), (2,)),
        0,
    )


def thermal_relaxation_choi(t1s: Array, tphis: Array, duration: float) -> Choi:
    """
    Construct a multi-qubit thermal relaxation channel.

    :param t1s: (num_qubits,) array of t1s.
    :param tphis: (num_qubits,) array of tphis.
    :param duration: The duration (same for all qubits).
    :return: The Choi matrix for the multi-qubit thermal relaxation channel.
    """
    num_qubits = len(t1s)
    if num_qubits == 1:
        return _thermal_relaxation_choi_1q(t1s[0], tphis[0], duration)

    # Use Python loop with static dimensions (unrolls at compile time when num_qubits is static)
    result = _thermal_relaxation_choi_1q(t1s[0], tphis[0], duration)

    for i in range(1, num_qubits):
        single_qubit_choi = _thermal_relaxation_choi_1q(t1s[i], tphis[i], duration)
        result = tensor_choi(result, single_qubit_choi)

    return result


def depolarizing_channel_superoperator(depolarizing_prob: float, num_qubits: int) -> SuperOp:
    """
    Construct the superoperator for a multi-qubit depolarizing channel.

    :param depolarizing_prob: The depolarizing probability (between 0 and 1).
    :param num_qubits: The number of qubits.
    :return: The superoperator matrix for the depolarizing channel.
    """
    d = 2**num_qubits
    identity_super = jnp.eye(d * d, dtype=complex)
    # The channel is defined as rho -> (1-p) rho + p * Tr(rho) * I / d
    # The superoperator for rho -> Tr(rho) * I / d is |I>> <<I| / d
    vec_identity = jnp.ravel(jnp.eye(d, dtype=complex))
    max_mixed_super = jnp.outer(vec_identity, vec_identity.conj()) / d

    depolarizing_super_data = (1 - depolarizing_prob) * identity_super + depolarizing_prob * max_mixed_super
    dims = (tuple([2] * num_qubits), tuple([2] * num_qubits))
    return SuperOp.from_matrix(depolarizing_super_data, dims, 0)


@jax.custom_vjp
def fractional_unitary_power(unitary: Unitary, exponent: float) -> Unitary:
    """
    Compute U^exponent for a unitary matrix with gradient support.

    This function computes fractional powers of unitary matrices and supports
    automatic differentiation using a custom VJP rule.

    The forward pass uses eigendecomposition:
    U = V @ diag(λ) @ V^(-1), so U^exponent = V @ diag(λ^exponent) @ V^(-1)

    The backward pass uses the Fréchet derivative of matrix powers, computed
    via a Sylvester equation solver for efficiency.

    :param unitary: The unitary object.
    :param exponent: The exponent (e.g., 1/n for the n-th root).
    :return: The matrix U^exponent as a Unitary object.
    """
    eigvals, eigvecs = jnp.linalg.eig(unitary.matrix)
    fractional_eigvals = jnp.power(eigvals, exponent)
    result_data = eigvecs @ jnp.diag(fractional_eigvals) @ jnp.linalg.inv(eigvecs)
    return Unitary.from_matrix(result_data, unitary.dims, unitary.num_ensemble_dims)


def _fractional_unitary_power_fwd(unitary: Unitary, exponent: float):
    """Forward pass: compute U^exponent and save values for backward pass."""
    eigvals, eigvecs = jnp.linalg.eig(unitary.matrix)
    fractional_eigvals = jnp.power(eigvals, exponent)
    V_inv = jnp.linalg.inv(eigvecs)
    result_data = eigvecs @ jnp.diag(fractional_eigvals) @ V_inv
    result = Unitary.from_matrix(result_data, unitary.dims, unitary.num_ensemble_dims)
    # Save for backward pass
    return result, (eigvals, eigvecs, V_inv, exponent, unitary.dims, unitary.num_ensemble_dims)


def _fractional_unitary_power_bwd(residuals, g: Unitary):
    """Backward pass: compute VJP using Fréchet derivative of matrix power.

    For f(U) = U^α, the Fréchet derivative is:
    Df(U)[V] = ∫₀^α U^(α-t) V U^(t-1) dt

    We compute this using a more efficient formulation with eigenvalues.
    """
    eigvals, eigvecs, V_inv, exponent, dims, num_ensemble_dims = residuals

    # g is the gradient w.r.t. output U^α, as a Unitary object
    # We need to compute gradient w.r.t. input U

    # Transform g to eigenspace: G = V^(-1) @ g @ V
    G = V_inv @ g.matrix @ eigvecs

    # Compute gradient in eigenspace using Loewner matrix
    # For diagonal matrices, D[diag(λ)^α] @ diag(δλ) involves (λ_i^α - λ_j^α)/(λ_i - λ_j)
    λ = eigvals
    λα = jnp.power(λ, exponent)

    # Loewner matrix: L[i,j] = (λ_i^α - λ_j^α) / (λ_i - λ_j) if i≠j, else α*λ_i^(α-1)
    λ_diff = λ[:, None] - λ[None, :]
    λα_diff = λα[:, None] - λα[None, :]

    # Avoid division by zero: use limit formula for i=j
    safe_denom = jnp.where(jnp.abs(λ_diff) < 1e-8, 1.0, λ_diff)
    L = jnp.where(
        jnp.abs(λ_diff) < 1e-8,
        exponent * jnp.power(λ[:, None], exponent - 1),  # Limit as λ_i -> λ_j
        λα_diff / safe_denom,  # type: ignore[operator]
    )

    # Gradient in eigenspace: element-wise multiplication
    G_eigen = L * G

    # Transform back to original space
    grad_unitary_data = eigvecs @ G_eigen @ V_inv
    grad_unitary = Unitary.from_matrix(grad_unitary_data, dims, num_ensemble_dims)

    return (grad_unitary, None)  # None for exponent (not differentiable)


# Register the custom VJP
fractional_unitary_power.defvjp(_fractional_unitary_power_fwd, _fractional_unitary_power_bwd)


# Wrap with JIT
fractional_unitary_power = partial(jit, static_argnames=("exponent",))(fractional_unitary_power)  # type: ignore[assignment]


def integrated_thermal_superoperator(
    unitary: Unitary,
    t1s: Array,
    tphis: Array,
    duration: float,
    num_steps: int = 100,
) -> SuperOp:
    """
    Construct a thermal channel which is integrated over a unitary rotation.

    For example, we may know accurate decoherence times for a qubit, and wish to calculate
    the channel over a unitary rotation by integrating the thermal relaxation during the gate.

    :param unitary: The unitary rotation object.
    :param t1s: (num_qubits,) array of T1 relaxation times.
    :param tphis: (num_qubits,) array of pure dephasing times.
    :param duration: The gate duration.
    :param num_steps: The number of steps to integrate over (static argument for JIT).
    :return: The superoperator integrated over the unitary rotation.
    """

    # 1. Compute fractional unitary: U^(1/num_steps)
    dU = fractional_unitary_power(unitary, 1.0 / num_steps)

    # Convert fractional unitary to superoperator
    dU_super = unitary_to_superop(dU)

    # 2. Compute thermal relaxation for fractional duration
    fractional_duration = duration / num_steps
    thermal_choi = thermal_relaxation_choi(t1s, tphis, fractional_duration)
    thermal_super = choi_to_superop(thermal_choi)

    # 3. Compute one step: thermal @ unitary
    superstep = compose_superop(thermal_super, dU_super)

    # 4. Compute the integrated channel by repeated matrix multiplication
    # Result = (S_thermal @ S_U)^num_steps
    # Use lax.scan for efficient repeated composition
    def scan_fn(acc, _):
        """Scan function: repeatedly apply the superstep."""
        return compose_superop(superstep, acc), None

    # Apply superstep (num_steps) times
    result, _ = jax.lax.scan(scan_fn, superstep, None, length=num_steps - 1)

    return result


def bit_flip_operators(p: float) -> tuple[Kraus, Kraus]:
    """Generate Kraus operators for a bit flip channel.

    The bit flip channel applies X with probability p and I with probability 1-p.

    :param p: Probability of bit flip error (0 <= p <= 1)
    :return: Tuple of two 2x2 Kraus operators (K0, K1)
    """
    return (
        Kraus.from_matrix(jnp.sqrt(1.0 - p) * I.matrix, ((2,), (2,)), 0),
        Kraus.from_matrix(jnp.sqrt(p) * X.matrix, ((2,), (2,)), 0),
    )


def phase_flip_operators(p: float) -> tuple[Kraus, Kraus]:
    """Generate Kraus operators for a phase flip channel.

    The phase flip channel applies Z with probability p and I with probability 1-p.

    :param p: Probability of phase flip error (0 <= p <= 1)
    :return: Tuple of two 2x2 Kraus operators (K0, K1)
    """
    return (
        Kraus.from_matrix(jnp.sqrt(1.0 - p) * I.matrix, ((2,), (2,)), 0),
        Kraus.from_matrix(jnp.sqrt(p) * Z.matrix, ((2,), (2,)), 0),
    )


def bitphase_flip_operators(p: float) -> tuple[Kraus, Kraus]:
    """Generate Kraus operators for a bit-phase flip channel.

    The bit-phase flip channel applies Y with probability p and I with probability 1-p.

    :param p: Probability of bit-phase flip error (0 <= p <= 1)
    :return: Tuple of two 2x2 Kraus operators (K0, K1)
    """
    return (
        Kraus.from_matrix(jnp.sqrt(1.0 - p) * I.matrix, ((2,), (2,)), 0),
        Kraus.from_matrix(jnp.sqrt(p) * Y.matrix, ((2,), (2,)), 0),
    )


def dephasing_operators(p: float) -> tuple[Kraus, Kraus]:
    """Generate Kraus operators for a dephasing (phase damping) channel.

    The dephasing channel causes loss of quantum information without energy dissipation.

    :param p: Dephasing probability (0 <= p <= 1)
    :return: Tuple of two 2x2 Kraus operators (K0, K1)
    """
    sqrt_p2 = jnp.sqrt(p / 2.0)
    sqrt_1mp2 = jnp.sqrt(1.0 - p / 2.0)
    return (
        Kraus.from_matrix(sqrt_1mp2 * I.matrix, ((2,), (2,)), 0),
        Kraus.from_matrix(sqrt_p2 * Z.matrix, ((2,), (2,)), 0),
    )


def depolarizing_operators(p: float) -> tuple[Kraus, Kraus, Kraus, Kraus]:
    """Generate Kraus operators for a depolarizing channel.

    The depolarizing channel applies I, X, Y, or Z each with equal probability p/4,
    and I with probability 1-p.

    :param p: Depolarizing probability (0 <= p <= 1)
    :return: Tuple of four 2x2 Kraus operators (K0, K1, K2, K3)
    """
    return (
        Kraus.from_matrix(jnp.sqrt(1.0 - p) * I.matrix, ((2,), (2,)), 0),
        Kraus.from_matrix(jnp.sqrt(p / 3.0) * X.matrix, ((2,), (2,)), 0),
        Kraus.from_matrix(jnp.sqrt(p / 3.0) * Y.matrix, ((2,), (2,)), 0),
        Kraus.from_matrix(jnp.sqrt(p / 3.0) * Z.matrix, ((2,), (2,)), 0),
    )


def relaxation_operators(p: float) -> tuple[Kraus, Kraus]:
    """Generate Kraus operators for an amplitude damping (relaxation) channel.

    The amplitude damping channel models energy dissipation (T1 decay).

    :param p: Relaxation probability (0 <= p <= 1)
    :return: Tuple of two 2x2 Kraus operators (K0, K1)
    """
    return (
        Kraus.from_matrix(jnp.array([[1.0, 0.0], [0.0, jnp.sqrt(1.0 - p)]], dtype=complex), ((2,), (2,)), 0),
        Kraus.from_matrix(jnp.array([[0.0, jnp.sqrt(p)], [0.0, 0.0]], dtype=complex), ((2,), (2,)), 0),
    )


# Dictionary mapping channel names to their Kraus operator functions
KRAUS_OPS = {
    "bit_flip": bit_flip_operators,
    "phase_flip": phase_flip_operators,
    "bitphase_flip": bitphase_flip_operators,
    "dephasing": dephasing_operators,
    "depolarizing": depolarizing_operators,
    "relaxation": relaxation_operators,
}



# Copyright 2026 Rigetti & Co, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce, singledispatch
from operator import mul
from typing import Tuple

import jax
import jax.numpy as jnp
from jax import Array

from ._quantum_objects import Choi, DensityMatrix, Kraus, KrausMap, PauliLiouville, StateVector, SuperOp, Unitary
from ._superoperator_transformations import choi_to_superop, pauli_liouville_to_superop


@jax.jit
def apply_superop_to_density_matrix(superop: SuperOp, rho: DensityMatrix) -> DensityMatrix:
    """Apply a superoperator to a density matrix.

    Supports ensemble broadcasting: superop and rho can have different ensemble sizes,
    and standard NumPy broadcasting rules apply.

    :param superop: Superoperator with shape ensemble_size + (d², d²)
    :param rho: Density matrix with shape ensemble_size + (d, d)
    :return: Transformed density matrix with broadcasted ensemble_size + (d, d)
    """
    assert superop.dims[0] == rho.dims, "Superoperator and density matrix must have the same dims."

    # Get dimension
    d = rho.d  # Linear dimension
    d2 = rho.d2  # Hilbert-Schmidt dimension

    # Get matrix representations
    superop_mat = superop.matrix
    rho_mat = rho.matrix

    # vec_col(rho) via transpose-last-axes then reshape
    rho_vec = jnp.reshape(jnp.swapaxes(rho_mat, -1, -2), rho.ensemble_size + (d2,))

    # Apply superoperator: vec(ρ_out) = S @ vec(ρ_in)
    # Use einsum with ellipsis for automatic broadcasting
    rho_out_vec = jnp.einsum("...ij,...j->...i", superop_mat, rho_vec)

    # Reshape back to matrix form (un-vectorize)
    # Column-stacking means we need to transpose back
    ensemble_size = jnp.broadcast_shapes(superop.ensemble_size, rho.ensemble_size)
    num_ensemble_dims = len(ensemble_size)

    # Unvec_col: reshape to (..., d, d) then transpose back
    rho_out = jnp.reshape(rho_out_vec, ensemble_size + (d, d))
    rho_out = jnp.swapaxes(rho_out, -1, -2)

    return DensityMatrix.from_matrix(rho_out, rho.dims, num_ensemble_dims)


@jax.jit
def apply_choi_to_density_matrix(choi: Choi, rho: DensityMatrix) -> DensityMatrix:
    """Apply a Choi matrix to a density matrix.

    Supports ensemble broadcasting: choi and rho can have different ensemble sizes,
    and standard NumPy broadcasting rules apply.

    :param choi: Choi matrix with shape ensemble_size + (d², d²)
    :param rho: Density matrix with shape ensemble_size + (d, d)
    :return: Transformed density matrix with broadcasted ensemble_size + (d, d)
    """
    # Convert to superoperator and apply
    superop = choi_to_superop(choi)
    return apply_superop_to_density_matrix(superop, rho)


@jax.jit
def apply_kraus_to_density_matrix(kraus_map: KrausMap, rho: DensityMatrix) -> DensityMatrix:
    """Apply a Kraus map to a density matrix.

    Computes E(ρ) = Σ_i K_i ρ K_i†

    Supports ensemble broadcasting: kraus_map and rho can have different ensemble sizes,
    and standard NumPy broadcasting rules apply.

    :param kraus_map: Kraus operators with shape ensemble_size + (n_kraus, d, d)
    :param rho: Density matrix with shape ensemble_size + (d, d)
    :return: Transformed density matrix with broadcasted ensemble_size + (d, d)
    """
    assert kraus_map.dims[0] == rho.dims, "Kraus map and density matrix must have the same dims."

    # Get matrix representations
    kraus_mat = kraus_map.matrix
    rho_mat = rho.matrix

    # Apply each Kraus operator: K_i @ rho @ K_i†
    # Use einsum with ellipsis for broadcasting
    # kraus_mat has shape (..., n_kraus, d, d)
    # rho_mat has shape (..., d, d)
    # Result: sum over i of K[..., i, :, :] @ rho[..., :, :] @ K[..., i, :, :].conj().T

    rho_out = jnp.einsum("...iab,...bc,...idc->...ad", kraus_mat, rho_mat, kraus_mat.conj())

    ensemble_size = jnp.broadcast_shapes(kraus_map.ensemble_size, rho.ensemble_size)
    num_ensemble_dims = len(ensemble_size)

    return DensityMatrix.from_matrix(rho_out, rho.dims, num_ensemble_dims)


@jax.jit
def apply_pauli_liouville_to_density_matrix(pl: PauliLiouville, rho: DensityMatrix) -> DensityMatrix:
    """Apply a Pauli Liouville representation to a density matrix.

    Supports ensemble broadcasting: pl and rho can have different ensemble sizes,
    and standard NumPy broadcasting rules apply.

    :param pl: Pauli-Liouville matrix with shape ensemble_size + (d², d²)
    :param rho: Density matrix with shape ensemble_size + (d, d)
    :return: Transformed density matrix with broadcasted ensemble_size + (d, d)
    """
    # Convert to superoperator and apply
    superop = pauli_liouville_to_superop(pl)
    return apply_superop_to_density_matrix(superop, rho)


@jax.jit
def apply_unitary_to_state_vector(unitary: Unitary, state: StateVector) -> StateVector:
    """Apply a unitary operator to a state vector.

    Computes |ψ_out⟩ = U |ψ_in⟩

    Supports ensemble broadcasting: unitary and state can have different ensemble sizes,
    and standard NumPy broadcasting rules apply.

    :param unitary: Unitary operator with shape ensemble_size + (d, d)
    :param state: State vector with shape ensemble_size + (d,)
    :return: Transformed state vector with broadcasted ensemble_size + (d,)
    """
    assert unitary.dims[0] == state.dims, "Unitary and state vector must have the same dims."

    # Get matrix representations
    unitary_mat = unitary.matrix
    state_mat = state.matrix

    # Apply unitary: |ψ_out⟩ = U |ψ_in⟩
    # Use einsum with ellipsis for broadcasting
    state_out = jnp.einsum("...ab,...b->...a", unitary_mat, state_mat)

    ensemble_size = jnp.broadcast_shapes(unitary.ensemble_size, state.ensemble_size)
    num_ensemble_dims = len(ensemble_size)

    return StateVector.from_matrix(state_out, state.dims, num_ensemble_dims)


@jax.jit
def apply_kraus_to_state_vector(kraus_map: Kraus, state: StateVector) -> DensityMatrix:
    """Apply a Kraus map to a state vector, resulting in a density matrix."""
    raise NotImplementedError("Applying Kraus operators to state vectors is not yet implemented.")


@singledispatch
def partial_trace(rho, indices: Tuple[int, ...]) -> DensityMatrix:
    raise TypeError("rho must be a DensityMatrix or Choi.")


@partial_trace.register
def _(rho: DensityMatrix, indices: Tuple[int, ...]) -> DensityMatrix:
    dims = tuple(rho.dims)  # e.g. (2,2)
    keep = tuple(indices)

    out_data = _partial_trace_data(rho.matrix, dims=dims, keep=keep)
    out_dims = tuple(dims[i] for i in sorted(keep))
    return DensityMatrix.from_matrix(out_data, out_dims, rho.num_ensemble_dims)


@partial_trace.register
def _(rho: Choi, indices: Tuple[int, ...]) -> DensityMatrix:
    dims_out, dims_in = rho.dims  # e.g. ((2,2),(2,2))
    dims_all = tuple(dims_out) + tuple(dims_in)  # e.g. (2,2,2,2)
    keep = tuple(indices)

    out_data = _partial_trace_data(rho.matrix, dims=dims_all, keep=keep)
    out_dims = tuple(dims_all[i] for i in sorted(keep))

    # After tracing arbitrary subsystems, "input vs output" split may not be meaningful,
    # so store dims as a flat tuple by default.
    return DensityMatrix.from_matrix(out_data, out_dims, rho.num_ensemble_dims)


@jax.jit(static_argnames=("dims", "keep"))
def _partial_trace_data(data: Array, dims: Tuple[int, ...], keep: Tuple[int, ...]) -> Array:
    """
    Batched partial trace of a matrix.

    data: (..., D, D)
    dims: (d0, d1, ..., dn-1) with prod(dims) == D
    keep: indices of subsystems to keep (in [0, n-1])

    returns: (..., D_keep, D_keep)
    """
    ensemble_shape = data.shape[:-2]
    D = data.shape[-1]
    n = len(dims)

    if reduce(mul, dims, 1) != D:
        raise ValueError(f"prod(dims)={reduce(mul, dims, 1)} must equal matrix dim D={D}.")

    if len(keep) != len(set(keep)):
        raise ValueError("Duplicate indices in keep.")
    if any((k < 0 or k >= n) for k in keep):
        raise IndexError(f"keep must be in [0, {n - 1}].")

    keep = tuple(sorted(keep))
    trace = tuple(i for i in range(n) if i not in keep)

    if len(trace) == 0:
        return data

    d_keep = reduce(mul, (dims[i] for i in keep), 1)
    d_trace = reduce(mul, (dims[i] for i in trace), 1)

    # reshape to (..., dims_row..., dims_col...)
    data_nd = data.reshape(ensemble_shape + dims + dims)

    # permute to (..., keep_row, trace_row, keep_col, trace_col)
    b = len(ensemble_shape)
    perm = (
        tuple(range(b))
        + tuple(b + i for i in keep)
        + tuple(b + i for i in trace)
        + tuple(b + n + i for i in keep)
        + tuple(b + n + i for i in trace)
    )
    data_perm = jnp.transpose(data_nd, perm)

    # group into (..., d_keep, d_trace, d_keep, d_trace)
    data_grp = data_perm.reshape(ensemble_shape + (d_keep, d_trace, d_keep, d_trace))

    # trace over traced subsystem
    out = jnp.trace(data_grp, axis1=-3, axis2=-1)  # (..., d_keep, d_keep)
    return out


@jax.jit
def compute_kraus_observables_from_states(
    kraus_map: KrausMap, input_states: DensityMatrix, observables: Unitary
) -> Array:
    """
    Compute the provided observables for the given input density matrices and process.

    :param kraus_ops: A Kraus Channel.
    :param input_states: A (num_states, d, d) array of density matrices.
    :param observables: A (num_observables, d, d) array of observables.
    :return (num_states, num_observables) array of expectation values.
    """
    # Get matrix representations
    input_mat = input_states.matrix
    obs_mat = observables.matrix
    kraus_mat = kraus_map.matrix

    # Compute obs: Tr[ρ Σ K_i† O K_i] = Σ Tr[K_i ρ K_i† O] -> ikj, kl, ilh
    # Kraus ops K†, K (num_kraus, d, d) -> ikj, ilh
    # Observable (num_observables, d, d) -> mkl
    # Input states 𝜌 (num_states, d, d) -> nhj
    # Result (num_states, num_observable) -> nm
    assert len(input_mat.shape) == 3, "input_states must be a (num_states, d, d) array"
    assert len(obs_mat.shape) == 3, "observables must be a (num_observables, d, d) array"
    assert len(kraus_mat.shape) == 3, "kraus_map must be a (num_kraus, d, d) array"
    predicted_expectations = jnp.real(
        jnp.einsum("nhj,ikj,mkl,ilh->nm", input_mat, kraus_mat.conj(), obs_mat, kraus_mat)
    )
    return predicted_expectations


@jax.jit
def compute_choi_observables_from_states(choi: Choi, input_states: DensityMatrix, observables: Unitary) -> Array:
    """
    Compute the provided observables for the given input density matrices and process.

    Uses the relation: Tr[O · E(ρ)] = vec(O)† · S · vec(ρ)
    where S is the superoperator (converted from Choi), E is the channel,
    ρ is the input state, O is the observable, and vec() is column-stacking vectorization.

    :param choi: A Choi matrix object.
    :param input_states: A (num_states, d, d) array of density matrices.
    :param observables: A (num_observables, d, d) array of observables.
    :return: A (num_states, num_observables) array of expectation values.
    """
    # Get matrix representations
    input_mat = input_states.matrix
    obs_mat = observables.matrix
    choi_mat = choi.matrix

    assert len(input_mat.shape) == 3, "input_states must be a (num_states, d, d) array"
    assert len(obs_mat.shape) == 3, "observables must be a (num_observables, d, d) array"
    assert len(choi_mat.shape) == 2, "choi must be a (d², d²) array"

    # Convert Choi to superoperator using existing function
    superop = choi_to_superop(choi)

    # Use the superoperator implementation
    return compute_superop_observables_from_states(superop, input_states, observables)


@jax.jit
def compute_superop_observables_from_states(
    superop: SuperOp, input_states: DensityMatrix, observables: Unitary
) -> Array:
    """
    Compute the provided observables for the given input density matrices and process.

    Uses the relation: Tr[O · E(ρ)] = vec(O)† · S · vec(ρ)
    where S is the superoperator, E is the channel, ρ is the input state,
    O is the observable, and vec() is column-stacking vectorization.

    :param superop: A superoperator matrix object.
    :param input_states: A (num_states, d, d) array of density matrices.
    :param observables: A (num_observables, d, d) array of observables.
    :return: A (num_states, num_observables) array of expectation values.
    """
    # Get matrix representations
    input_mat = input_states.matrix
    obs_mat = observables.matrix
    superop_mat = superop.matrix

    assert len(input_mat.shape) == 3, "input_states must be a (num_states, d, d) array"
    assert len(obs_mat.shape) == 3, "observables must be a (num_observables, d, d) array"
    assert len(superop_mat.shape) == 2, "superop must be a (d², d²) array"

    # Get dimension from superoperator matrix
    d2 = superop.d2[0]

    # Vectorize states and observables using column-stacking convention
    # vec(M)[i*d+j] = M[j,i] for column-stacking
    # So M.T.reshape(-1) gives the column-stacked vector
    vec_states = input_mat.transpose(0, 2, 1).reshape(-1, d2)  # (num_states, d²)
    vec_obs = obs_mat.transpose(0, 2, 1).reshape(-1, d2)  # (num_observables, d²)

    # Compute: Tr[O · E(ρ)] = vec(O)† · S · vec(ρ)
    # result[n,m] = Σ_rs vec_obs[m,r]^* · superop[r,s] · vec_states[n,s]
    predicted_expectations = jnp.real(
        vec_obs.conj() @ superop_mat @ vec_states.T
    ).T  # Transpose to get (num_states, num_observables)

    return predicted_expectations


@jax.jit
def compute_pauli_liouville_observables_from_states(
    pauli_liouville: PauliLiouville, input_states: DensityMatrix, observables: Unitary
) -> Array:
    """
    Compute the provided observables for the given input density matrices and process.

    Converts the Pauli-Liouville representation to superoperator form, then uses:
    Tr[O · E(ρ)] = vec(O)† · S · vec(ρ)
    where S is the superoperator, E is the channel, ρ is the input state,
    O is the observable, and vec() is column-stacking vectorization.

    :param pauli_liouville: A Pauli-Liouville matrix object.
    :param input_states: A (num_states, d, d) array of density matrices.
    :param observables: A (num_observables, d, d) array of observables.
    :return: A (num_states, num_observables) array of expectation values.
    """
    # Get matrix representations
    input_mat = input_states.matrix
    obs_mat = observables.matrix
    pl_mat = pauli_liouville.matrix

    assert len(input_mat.shape) == 3, "input_states must be a (num_states, d, d) array"
    assert len(obs_mat.shape) == 3, "observables must be a (num_observables, d, d) array"
    assert len(pl_mat.shape) == 2, "pauli_liouville must be a (d², d²) array"

    # Convert Pauli-Liouville to superoperator
    superop = pauli_liouville_to_superop(pauli_liouville)

    # Use the superoperator implementation
    return compute_superop_observables_from_states(superop, input_states, observables)



# Copyright 2026 Rigetti & Co, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
from jax.numpy import exp, pi, sqrt

from ._quantum_objects import StateVector

KET0 = ZPLUS = StateVector.from_matrix(jnp.array([1, 0], dtype=complex), (2,), 0)
"""|0> state vector."""

KET1 = ZMINUS = StateVector.from_matrix(jnp.array([0, 1], dtype=complex), (2,), 0)
"""|1> state vector."""

KETPLUS = XPLUS = StateVector.from_matrix(jnp.array([1, 1], dtype=complex) / sqrt(2), (2,), 0)
"""|+> = (|0> + |1>) / sqrt(2) state vector."""

KETMINUS = XMINUS = StateVector.from_matrix(jnp.array([1, -1], dtype=complex) / sqrt(2), (2,), 0)
"""|-> = (|0> - |1>) / sqrt(2) state vector."""

KETPLUSI = YPLUS = StateVector.from_matrix(jnp.array([1, 1j], dtype=complex) / sqrt(2), (2,), 0)
"""|+i> = (|0> + i|1>) / sqrt(2) state vector."""

KETMINUSI = YMINUS = StateVector.from_matrix(jnp.array([1, -1j], dtype=complex) / sqrt(2), (2,), 0)
"""|-i> = (|0> - i|1>) / sqrt(2) state vector."""


SIC0 = KET0
SIC1 = StateVector.from_matrix(
    exp(1j * pi / 2) * (1 / sqrt(3) * KET0.matrix) + exp(1j * pi / 2) * sqrt(2 / 3) * KET1.matrix, (2,), 0
)
SIC2 = StateVector.from_matrix(
    exp(1j * 5 * pi / 6) * (1 / sqrt(3) * KET0.matrix) + exp(1j * 1 * pi / 6) * sqrt(2 / 3) * KET1.matrix, (2,), 0
)
SIC3 = StateVector.from_matrix(
    exp(1j * pi / 6) * (1 / sqrt(3) * KET0.matrix) + exp(1j * 5 * pi / 6) * sqrt(2 / 3) * KET1.matrix, (2,), 0
)

SIC_STATES = {
    "SIC0": SIC1,
    "SIC1": SIC1,
    "SIC2": SIC2,
    "SIC3": SIC3,
}
"""
The symmetric informationally complete POVMs for a qubit.

These can reduce the number of experiments to perform quantum process tomography.
For more information, please see http://info.phys.unm.edu/~caves/reports/infopovm.pdf
"""

PAULI_STATES = {
    "X+": XPLUS,
    "X-": XMINUS,
    "Y+": YPLUS,
    "Y-": YMINUS,
    "Z+": ZPLUS,
    "Z-": ZMINUS,
}
"""The six eigenstates of the Pauli operators X, Y, and Z."""

STATES = {
    "X": [
        StateVector.from_matrix(jnp.array([1, 1], dtype=complex) / jnp.sqrt(2), (2,), 0),
        StateVector.from_matrix(jnp.array([1, -1], dtype=complex) / jnp.sqrt(2), (2,), 0),
    ],
    "Y": [
        StateVector.from_matrix(jnp.array([1, 1j], dtype=complex) / jnp.sqrt(2), (2,), 0),
        StateVector.from_matrix(jnp.array([1, -1j], dtype=complex) / jnp.sqrt(2), (2,), 0),
    ],
    "Z": [
        StateVector.from_matrix(jnp.array([1, 0], dtype=complex), (2,), 0),
        StateVector.from_matrix(jnp.array([0, 1], dtype=complex), (2,), 0),
    ],
    "SIC": [SIC0, SIC1, SIC2, SIC3],
}
